from __future__ import annotations

import shutil
import subprocess
from copy import deepcopy
from pathlib import Path
from typing import Mapping, NotRequired, TypedDict

from loguru import logger

from statickg.helper import (
    find_available_port,
    get_classpath,
    is_port_available,
    wait_till_port_available,
)
from statickg.models.prelude import (
    ETLOutput,
    RelPathRefStr,
    RelPathRefStrOrStr,
    Repository,
)
from statickg.services.fuseki import DBInfo, FusekiDataLoaderService
from statickg.services.interface import BaseService


class FusekiDeploymentServiceConstructArgs(TypedDict):
    hostname: NotRequired[str]


class FusekiDeploymentServiceInvokeArgs(TypedDict):
    start: RelPathRefStrOrStr
    stop_all: RelPathRefStrOrStr


class FusekiDeploymentService(BaseService[FusekiDeploymentServiceInvokeArgs]):

    def __init__(
        self,
        name: str,
        workdir: Path,
        args: FusekiDeploymentServiceConstructArgs,
        services: Mapping[str, BaseService],
    ):
        self.name = name
        self.workdir = workdir
        self.services = services
        self.logger = logger.bind(name=get_classpath(self.__class__).rsplit(".", 1)[0])
        self.args = args
        self.hostname = args.get("hostname", "http://localhost")
        self.current_dbinfo = None

    def forward(
        self,
        repo: Repository,
        args: FusekiDeploymentServiceInvokeArgs,
        agg_output: ETLOutput,
    ):
        args = deepcopy(args)
        lst: list[DBInfo] = agg_output.output[get_classpath(FusekiDataLoaderService)]
        (dbinfo,) = lst

        if not dbinfo.has_running_service():
            if not is_port_available(self.hostname, 3030):
                subprocess.check_call(self.get_stop_all_command(args), shell=True)

            if not wait_till_port_available(self.hostname, 3030, timeout=10):
                self.logger.error(
                    "After stopping all fuseki services, port 3030 is still not available."
                )
                raise Exception(
                    "Another one started a service on port 3030 that is not managed by this service"
                )

            # only deploy the service when there is no running service for the directory
            if self.start_fuseki(args, dbinfo):
                # remove old databases
                for old_db in dbinfo.get_older_versions():
                    shutil.rmtree(old_db.dir)

        return

    def start_fuseki(self, args: FusekiDeploymentServiceInvokeArgs, dbinfo: DBInfo):
        if dbinfo.hostname is not None:
            return False

        name = f"fuseki-{dbinfo.dir.name}"
        port = find_available_port(self.hostname, 3030)
        try:
            subprocess.check_call(
                self.get_start_command(args).format(
                    ID=name, PORT=str(port), DB_DIR=dbinfo.dir
                ),
                shell=True,
            )
        except subprocess.CalledProcessError:
            subprocess.check_call(
                self.get_stop_all_command(args).format(ID=name),
                shell=True,
            )
            subprocess.check_call(
                self.get_start_command(args).format(
                    ID=name, PORT=str(port), DB_DIR=dbinfo.dir
                ),
                shell=True,
            )

        dbinfo.hostname = f"{self.hostname}:{port}"
        self.logger.debug(
            "Started Fuseki service at {} serving {}", dbinfo.hostname, dbinfo.dir.name
        )
        return True

    def get_start_command(self, args: FusekiDeploymentServiceInvokeArgs):
        cmd = args["start"]
        if isinstance(cmd, RelPathRefStr):
            cmd = cmd.deref()
            # trick to avoid calling deref() again
            args["start"] = cmd
        return cmd

    def get_stop_all_command(self, args: FusekiDeploymentServiceInvokeArgs):
        cmd = args["stop_all"]
        if isinstance(cmd, RelPathRefStr):
            cmd = cmd.deref()
            # trick to avoid calling deref() again
            args["stop_all"] = cmd
        return cmd



from __future__ import annotations

import re
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Literal, Optional

import numpy as np
import strsim
from rdflib import RDFS, SKOS, Graph, URIRef


@dataclass
class Doc:
    id: str
    labels: list[str]
    props: dict[str, str]


class EntityLinking:
    instances = {}

    def __init__(self, data_file: Path | str, format: str):
        self.data_file = Path(data_file)
        self.g = Graph()
        self.g.parse(self.data_file, format=format)

        self.docs: list[Doc] = []
        for subj in self.g.subjects(unique=True):
            if not isinstance(subj, URIRef):
                continue
            labels = [str(x) for x in self.g.objects(subj, RDFS.label, unique=True)]
            labels.extend(
                (str(x) for x in self.g.objects(subj, SKOS.altLabel, unique=True))
            )
            if len(labels) == 0:
                continue

            self.docs.append(
                Doc(
                    id=str(subj),
                    labels=labels,
                    props={
                        str(pred): str(obj)
                        for pred, obj in self.g.predicate_objects(subj)
                        if pred != RDFS.label
                    },
                )
            )
        self.id2doc = {doc.id: doc for doc in self.docs}
        self.feat_extractor = FeatExtractor()

    @staticmethod
    def get_instance(
        entity_dir: Path | str,
        name: Literal[
            "crs",
            "country",
            "state_or_province",
            "commodity",
            "unit",
            "material_form",
            "category",
        ],
    ) -> EntityLinking:
        entity_dir = Path(entity_dir)
        if name not in EntityLinking.instances:
            mno = "https://minmod.isi.edu/ontology/"
            mnr = "https://minmod.isi.edu/resource/"
            if name == "crs":
                linker = EntityLinking(entity_dir / "epsg.ttl", "turtle")
            elif name == "state_or_province":
                linker = EntityLinking(entity_dir / "state_or_province.ttl", "turtle")
                country_linker = EntityLinking.get_instance(entity_dir, "country")
                name2country = {doc.labels[0]: doc for doc in country_linker.docs}
                for doc in linker.docs:
                    doc.props[f"{mno}country"] = name2country[
                        doc.props[f"{mno}country"]
                    ].id
            elif name in ["commodity", "unit", "country", "material_form", "category"]:
                linker = EntityLinking(entity_dir / f"{name}.ttl", "turtle")
            else:
                raise ValueError(f"Unknown entity type: {name}")

            EntityLinking.instances[name] = linker
        return EntityLinking.instances[name]

    def link(
        self, query: str, has_props: Optional[dict[str, str]] = None
    ) -> Optional[tuple[Doc, float]]:
        if has_props is None:
            has_props = {}

        scores = [
            (i, self.feat_extractor.extract(query, doc.labels).mean())
            for i, doc in enumerate(self.docs)
            if all(doc.props.get(k) == v for k, v in has_props.items())
        ]

        if len(scores) == 0:
            return None

        i, score = max(scores, key=lambda x: x[1])
        return self.docs[i], float(score)


class FeatExtractor:
    def __init__(self):
        self.chartok = strsim.CharacterTokenizer()
        self.charseqtok = strsim.WhitespaceCharSeqTokenizer()

    def extract(self, text: str, labels: Iterable[str]) -> np.ndarray:
        feat = np.zeros((7,), dtype=np.float64)
        for label in labels:
            feat = np.maximum(feat, self.extract_pairwise_features_v2(text, label))
        return feat

    def extract_pairwise_features_v2(self, text: str, entity_label: str):
        text_t1 = self.chartok.tokenize(text)
        entity_label_t1 = self.chartok.tokenize(entity_label)

        text_t2 = self.charseqtok.tokenize(text)
        entity_label_t2 = self.charseqtok.tokenize(entity_label)

        text_t3 = self.charseqtok.unique_tokenize(text)
        entity_label_t3 = self.charseqtok.unique_tokenize(entity_label)

        out2 = [
            strsim.levenshtein_similarity(text_t1, entity_label_t1),
            strsim.jaro_winkler_similarity(text_t1, entity_label_t1),
            strsim.monge_elkan_similarity(text_t2, entity_label_t2),
            (
                sym_monge_score := strsim.symmetric_monge_elkan_similarity(
                    text_t2, entity_label_t2
                )
            ),
            (hyjac_score := strsim.hybrid_jaccard_similarity(text_t3, entity_label_t3)),
            does_ordinal_match(text, entity_label, sym_monge_score, 0.7),
            does_ordinal_match(text, entity_label, hyjac_score, 0.7),
        ]
        return out2


def does_ordinal_match(s1: str, s2: str, sim: float, threshold: float) -> float:
    """Test for strings containing ordinal categorical values such as Su-30 vs Su-25, 29th Awards vs 30th Awards.

    Args:
        s1: Cell Label
        s2: Entity Label
    """
    if sim < threshold:
        return 0.4

    digit_tokens_1 = re.findall(r"\d+", s1)
    digit_tokens_2 = re.findall(r"\d+", s2)

    if digit_tokens_1 == digit_tokens_2:
        return 1.0

    if len(digit_tokens_1) == 0 or len(digit_tokens_2) == 0:
        return 0.4

    return 0.0


if __name__ == "__main__":
    pass



from __future__ import annotations

from datetime import datetime
from typing import Any, Optional, TypeVar

import pandas as pd
import requests
from shapely.errors import WKTReadingError
from shapely.geometry import GeometryCollection
from shapely.wkt import dumps, loads

DEFAULT_ENDPOINT = "https://minmod.isi.edu/sparql"
MNR_NS = "https://minmod.isi.edu/resource/"
V = TypeVar("V")


class UnconvertibleUnitError(Exception):
    pass


def group_by_key(output: list[dict], key: str) -> dict[str, list[dict]]:
    groups = {}
    for row in output:
        val = row[key]
        if val not in groups:
            groups[val] = []
        groups[val].append(row)
    return groups


def group_by_attr(output: list[V], attr: str) -> dict[str, list[V]]:
    groups = {}
    for row in output:
        val = getattr(row, attr)
        if val not in groups:
            groups[val] = []
        groups[val].append(row)
    return groups


def send_sparql_query(query, endpoint=DEFAULT_ENDPOINT):
    final_query = (
        """
    PREFIX dcterms: <http://purl.org/dc/terms/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX : <https://minmod.isi.edu/ontology/>
    PREFIX mnr: <https://minmod.isi.edu/resource/>
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    PREFIX owl: <http://www.w3.org/2002/07/owl#>
    PREFIX gkbi: <https://geokb.wikibase.cloud/entity/>
    PREFIX gkbt: <https://geokb.wikibase.cloud/prop/direct/>
    PREFIX geo: <http://www.opengis.net/ont/geosparql#>
    PREFIX skos: <http://www.w3.org/2004/02/skos/core#>
    \n"""
        + query
    )
    # send query
    response = requests.post(
        url=endpoint,
        data={"query": final_query},
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/sparql-results+json",  # Requesting JSON format
        },
        verify=False,  # Set to False to bypass SSL verification as per the '-k' in curl
    )
    return response


def run_sparql_query(
    query,
    endpoint=DEFAULT_ENDPOINT,
    keys: Optional[list[str]] = None,
) -> list[dict]:
    response = send_sparql_query(query, endpoint)

    if response.status_code != 200:
        raise Exception(response.text)

    qres = response.json()["results"]["bindings"]

    if len(qres) == 0:
        return []

    sample = {}
    for res in qres:
        for key in res:
            if key not in sample:
                sample[key] = res[key]
    if keys is not None:
        for key in keys:
            if key not in sample:
                sample[key] = {"type": "bnode"}

    output = [{} for _ in range(len(qres))]
    for key, val in sample.items():
        if val["type"] in {"uri", "bnode"}:
            for i, row in enumerate(qres):
                if key in row:
                    output[i][key] = row[key]["value"]
                else:
                    output[i][key] = None
        else:
            assert val["type"] == "literal", val["type"]
            if val.get("datatype") in {
                None,
                "http://www.opengis.net/ont/geosparql#wktLiteral",
            }:
                for i, row in enumerate(qres):
                    if key in row:
                        output[i][key] = row[key]["value"]
                    else:
                        output[i][key] = None
            elif val["datatype"] in {
                "http://www.w3.org/2001/XMLSchema#integer",
            }:
                for i, row in enumerate(qres):
                    if key in row:
                        output[i][key] = int(row[key]["value"])
                    else:
                        output[i][key] = None
            elif val["datatype"] in {
                "http://www.w3.org/2001/XMLSchema#decimal",
                "http://www.w3.org/2001/XMLSchema#double",
            }:
                for i, row in enumerate(qres):
                    if key in row:
                        output[i][key] = float(row[key]["value"])
                    else:
                        output[i][key] = None
            elif val["datatype"] == "http://www.w3.org/2001/XMLSchema#date":
                for i, row in enumerate(qres):
                    if key in row:
                        rowval = row[key]["value"]
                        try:
                            output[i][key] = datetime.strptime(rowval, "%Y-%m-%d")
                        except ValueError:
                            output[i][key] = datetime.strptime(rowval, "%Y-%m")
                    else:
                        output[i][key] = None
            else:
                raise NotImplementedError(val)
    return output


def assert_isinstance(x: Any, cls: type[V]) -> V:
    assert isinstance(x, cls)
    return x


def merge_wkt(series):
    geometries = []
    for wkt in series:
        if pd.notna(wkt) and isinstance(wkt, str):
            try:
                geometry = loads(wkt)
                geometries.append(geometry)
            except Exception as e:
                print(f"Warning: Error loading WKT: {e} for WKT: {wkt}, skipping entry")

    if len(geometries) == 1:
        # return the single geometry directly
        return dumps(geometries[0])
    elif len(geometries) > 1:
        # return a GEOMETRYCOLLECTION if there are multiple geometries
        return dumps(GeometryCollection(geometries))
    else:
        # return None if there are no valid geometries
        return None



from __future__ import annotations

from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from functools import cached_property, cmp_to_key, lru_cache
from typing import Iterable, Optional, Union

from minmodkg.misc import (
    MNR_NS,
    UnconvertibleUnitError,
    V,
    assert_isinstance,
    group_by_attr,
)


class ResourceCategory(str, Enum):
    """Resource category of a mineral site"""

    Inferred = f"{MNR_NS}Inferred"
    Indicated = f"{MNR_NS}Indicated"
    Measured = f"{MNR_NS}Measured"


class ReserveCategory(str, Enum):
    """Reserve category of a mineral site"""

    Proven = f"{MNR_NS}Proven"
    Probable = f"{MNR_NS}Probable"


# note that they may have more such as accumulated, previously mined, etc.
ResourceReserveCategory = str
Mt_unit = f"{MNR_NS}Q202"
percent_unit = f"{MNR_NS}Q201"


@dataclass(frozen=True)
class GradeTonnageEstimate:
    tonnage: float
    contained_metal: float

    def is_equal_or_better(self, other: GradeTonnageEstimate):
        return self.contained_metal >= other.contained_metal

    def add(self, other: GradeTonnageEstimate) -> GradeTonnageEstimate:
        return GradeTonnageEstimate(
            self.tonnage + other.tonnage, self.contained_metal + other.contained_metal
        )


@dataclass(frozen=True)
class SiteGradeTonnage:
    resource_estimate: Optional[GradeTonnageEstimate] = None
    reserve_estimate: Optional[GradeTonnageEstimate] = None

    @cached_property
    def total_resource_contained_metal(self):
        if self.resource_estimate is not None:
            return self.resource_estimate.contained_metal
        return None

    @cached_property
    def total_resource_tonnage(self):
        if self.resource_estimate is not None:
            return self.resource_estimate.tonnage
        return None

    @cached_property
    def total_reserve_contained_metal(self):
        if self.reserve_estimate is not None:
            return self.reserve_estimate.contained_metal
        return None

    @cached_property
    def total_reserve_tonnage(self):
        if self.reserve_estimate is not None:
            return self.reserve_estimate.tonnage
        return None

    @cached_property
    def total_compared_contained_metal(self):
        if self.total_resource_contained_metal is not None:
            return self.total_resource_contained_metal
        return self.total_reserve_contained_metal

    @cached_property
    def total_contained_metal(self):
        total = 0
        if self.total_resource_contained_metal is not None:
            total += self.total_resource_contained_metal
        if self.total_reserve_contained_metal is not None:
            total += self.total_reserve_contained_metal
        return total

    def get_total_resource_grade(self, unit: Optional[str] = None) -> Optional[float]:
        if unit is None:
            unit = percent_unit

        total_resource_contained_metal = self.total_resource_contained_metal
        if total_resource_contained_metal is None:
            return None

        total_resource_tonnage = self.total_resource_tonnage
        assert total_resource_tonnage is not None
        if total_resource_contained_metal == 0.0:
            assert total_resource_tonnage == 0.0
            return 0.0

        return unit_conversion(
            total_resource_contained_metal / total_resource_tonnage * 100,
            percent_unit,
            unit,
        )

    def get_total_reserve_grade(self, unit: Optional[str] = None) -> Optional[float]:
        if unit is None:
            unit = percent_unit

        total_reserve_contained_metal = self.total_reserve_contained_metal
        if total_reserve_contained_metal is None:
            return None

        total_reserve_tonnage = self.total_reserve_tonnage
        assert total_reserve_tonnage is not None
        if total_reserve_contained_metal == 0.0:
            assert total_reserve_tonnage == 0.0
            return 0.0

        return unit_conversion(
            total_reserve_contained_metal / total_reserve_tonnage * 100,
            percent_unit,
            unit,
        )

    def is_equal_or_better(self, other: SiteGradeTonnage):
        if self.total_compared_contained_metal is not None:
            if other.total_compared_contained_metal is not None:
                return (
                    self.total_compared_contained_metal
                    >= other.total_compared_contained_metal
                )
            return True
        elif other.total_compared_contained_metal is not None:
            return False

        return self.total_contained_metal >= other.total_contained_metal

    def add(self, other: SiteGradeTonnage):
        resource_estimate = self.resource_estimate
        reserve_estimate = self.reserve_estimate

        if resource_estimate is not None:
            if other.resource_estimate is not None:
                resource_estimate = resource_estimate.add(other.resource_estimate)
        elif other.resource_estimate is not None:
            resource_estimate = other.resource_estimate

        if reserve_estimate is not None:
            if other.reserve_estimate is not None:
                reserve_estimate = reserve_estimate.add(other.reserve_estimate)
        elif other.reserve_estimate is not None:
            reserve_estimate = other.reserve_estimate
        return SiteGradeTonnage(resource_estimate, reserve_estimate)


class GradeTonnageModel:
    """Computing grade & tonnage data of a single mineral site from mineral inventories"""

    @dataclass
    class MineralInventory:
        id: str  # unique identifier of the inventory -- for knowing the reported data is sum of multiple categories or not
        date: Optional[
            str
        ]  # %YYYY-%MM-%DD: this allow us to group by and sort by date without parsing it
        zone: Optional[str]
        category: list[str]
        ore_value: float
        ore_unit: str
        grade_value: float
        grade_unit: str

    def __call__(
        self,
        invs: list[MineralInventory],
        norm_tonnage_unit: Optional[str] = None,
        norm_grade_unit: Optional[str] = None,
    ):
        if norm_tonnage_unit is None:
            norm_tonnage_unit = Mt_unit
        if norm_grade_unit is None:
            norm_grade_unit = percent_unit

        resource_cat = frozenset({c.value for c in ResourceCategory})
        reserve_cat = frozenset({c.value for c in ReserveCategory})

        # group by zone & date
        grade_tonnages = []
        for date, invs_by_date in group_by_attr(invs, "date").items():
            grade_tonnage_per_zones = []
            for zone, invs_by_date_zone in group_by_attr(invs_by_date, "zone").items():
                # the extraction may went wrong and we have multiple results per category
                # therefore, we need to handle them.
                # assert len(invs_by_date_zone) == 1

                # the first step is normalization
                cat2ests = defaultdict(list)
                for inv in invs_by_date_zone:
                    try:
                        ore = unit_conversion(
                            inv.ore_value, inv.ore_unit, norm_tonnage_unit
                        )
                        grade = unit_conversion(
                            inv.grade_value, inv.grade_unit, norm_grade_unit
                        )
                    except UnconvertibleUnitError as e:
                        # the data is broken, so we skip it
                        continue

                    cat = frozenset(inv.category)
                    if not (cat.issubset(resource_cat) or cat.issubset(reserve_cat)):
                        # ignore errorneous data
                        continue

                    cat2ests[frozenset(inv.category)].append(
                        GradeTonnageEstimate(
                            tonnage=ore,
                            contained_metal=ore
                            * unit_conversion(grade, norm_grade_unit, percent_unit)
                            / 100,
                        )
                    )

                cat_est = [
                    (
                        cat,
                        max(
                            ests,
                            key=cmp_to_key(GradeTonnageEstimate.is_equal_or_better),
                        ),
                    )
                    for cat, ests in cat2ests.items()
                ]

                # now, we need to compute resource/reserve estimates by summing up the estimate
                resource_est = [x for x in cat_est if x[0].issubset(resource_cat)]
                reserve_est = [x for x in cat_est if x[0].issubset(reserve_cat)]

                attr2est = {"resource": None, "reserve": None}
                for attr, ests in [
                    ("resource", resource_est),
                    ("reserve", reserve_est),
                ]:
                    allcats = {cat for cat, _ in ests}
                    while True:
                        new_ests = []
                        for i in range(len(ests)):
                            cat, est = ests[i]
                            for j in range(i + 1, len(ests)):
                                if cat.isdisjoint(ests[j][0]):
                                    newcat = cat.union(ests[j][0])
                                    if newcat not in allcats:
                                        # we can merge them
                                        new_ests.append((newcat, est.add(ests[j][1])))
                                        allcats.add(newcat)
                        if len(new_ests) == 0:
                            break
                        ests.extend(new_ests)

                    if len(ests) != 0:
                        attr2est[attr] = max(
                            (x[1] for x in ests),
                            key=cmp_to_key(GradeTonnageEstimate.is_equal_or_better),
                        )

                grade_tonnage_per_zones.append(
                    (
                        zone,
                        SiteGradeTonnage(
                            resource_estimate=attr2est["resource"],
                            reserve_estimate=attr2est["reserve"],
                        ),
                    )
                )

            if len(grade_tonnage_per_zones) == 0:
                continue

            grade_tonnages.append(
                (date, self.aggregate_site_tonnages_by_zone(grade_tonnage_per_zones))
            )

        if len(grade_tonnages) == 0:
            return None

        return self.aggregate_site_tonnages_by_date(grade_tonnages)

    def aggregate_site_tonnages_by_zone(
        self, vals: list[tuple[Optional[str], SiteGradeTonnage]]
    ):
        site_tonnage = None
        zone_tonnage = None

        for zone, val in vals:
            if zone is None:
                if site_tonnage is not None:
                    raise Exception(
                        "We should encounter the site inventory (no zone) only once"
                    )
                site_tonnage = val
            else:
                if zone_tonnage is None:
                    zone_tonnage = val
                else:
                    zone_tonnage = zone_tonnage.add(val)

        if site_tonnage is not None:
            if zone_tonnage is not None:
                if site_tonnage.is_equal_or_better(zone_tonnage):
                    return site_tonnage
                return zone_tonnage

            return site_tonnage
        else:
            assert zone_tonnage is not None
            return zone_tonnage

    def aggregate_site_tonnages_by_date(
        self, vals: Iterable[tuple[Optional[str], SiteGradeTonnage]]
    ):
        return max(vals, key=cmp_to_key(lambda a, b: a[1].is_equal_or_better(b[1])))[1]


weight_uncompatible_units = {
    f"{MNR_NS}{id}"
    for id in [
        "Q201",
        "Q203",
        "Q204",
        "Q205",
        "Q207",
        "Q208",
        "Q209",
        "Q210",
        "Q212",
        "Q216",
        "Q217",
        "Q220",
        "Q206",
        "Q211",
        "Q221",
        "Q218",
        "Q219",
    ]
}
percent_uncompatible_units = {
    f"{MNR_NS}{id}"
    for id in [
        "Q200",
        "Q202",
        "Q204",
        "Q205",
        "Q206",
        "Q207",
        "Q208",
        "Q209",
        "Q210",
        "Q211",
        "Q212",
        "Q213",
        "Q214",
        "Q215",
        "Q216",
        "Q218",
        "Q219",
        "Q221",
    ]
}


def unit_conversion(value: float, unit: str, to_unit: str) -> float:

    if unit == to_unit:
        return value

    if to_unit == f"{MNR_NS}Q202":
        # convert to million tonnes
        if unit == f"{MNR_NS}Q200":
            # from tonnes
            return value / 1_000_000
        if unit == f"{MNR_NS}Q213":
            # from million short tons
            return value / 1.10231
        if unit == f"{MNR_NS}Q214":
            return value / 1_000_000 / 1.10231
        if unit == f"{MNR_NS}Q215":
            # million pounds
            return value * 0.000454
        if unit in weight_uncompatible_units:
            raise UnconvertibleUnitError((value, unit, to_unit))
        raise NotImplementedError((value, unit, to_unit))

    if to_unit == f"{MNR_NS}Q201":
        # convert to percentage
        if unit == f"{MNR_NS}Q203" or unit == f"{MNR_NS}Q220":
            # from grams per tonne or parts per million
            return value / 10_000
        if unit == f"{MNR_NS}Q217":
            # from kg per tonne
            return value / 10
        if unit in percent_uncompatible_units:
            raise UnconvertibleUnitError((value, unit, to_unit))
        raise NotImplementedError((value, unit, to_unit))

    raise NotImplementedError((value, unit, to_unit))



from __future__ import annotations

import os
from functools import lru_cache
from typing import Annotated, Literal, Optional

import networkx as nx
import pandas as pd
from fastapi import APIRouter, FastAPI, Header, HTTPException, Response
from minmodkg.grade_tonnage_model import (
    GradeTonnageEstimate,
    GradeTonnageModel,
    ResourceCategory,
    SiteGradeTonnage,
)
from minmodkg.misc import group_by_key, merge_wkt, run_sparql_query

"""
An endpoint to allow querying derived data from the Minmod knowledge graph.
"""
app = FastAPI(openapi_url="/api/v1/openapi.json", docs_url="/api/v1/docs")
DEFAULT_ENDPOINT = os.environ.get("SPARQL_ENDPOINT", "https://minmod.isi.edu/sparql")
MNR_NS = "https://minmod.isi.edu/resource/"
router = APIRouter(
    prefix="/api/v1",
)


@router.get("/deposit_types")
def deposit_types():
    return get_deposit_types(get_snapshot_id())


@router.get("/commodities")
def commodities():
    return get_commodities(get_snapshot_id())


@router.get("/mineral_site_grade_and_tonnage/{commodity}")
def mineral_site_grade_and_tonnage(
    commodity: str,
    norm_tonnage_unit: Optional[str] = None,
    norm_grade_unit: Optional[str] = None,
    date_precision: Literal["year", "month", "day"] = "month",
    accept: Annotated[str | None, Header()] = None,
):
    commodity = norm_commodity(commodity)
    output = get_grade_tonnage_inventory(
        get_snapshot_id(),
        commodity,
        norm_tonnage_unit=norm_tonnage_unit,
        norm_grade_unit=norm_grade_unit,
        date_precision=date_precision,
    )
    if accept is not None and "text/csv" in accept:
        df = pd.DataFrame(output)
        return Response(
            df.to_csv(index=False, float_format="%.5f"), media_type="text/csv"
        )
    return output


@router.get("/mineral_site_deposit_types/{commodity}")
def mineral_site_deposit_types(
    commodity: str,
    accept: Annotated[str | None, Header()] = None,
):
    commodity = norm_commodity(commodity)
    output = get_deposit_type_classification(
        get_snapshot_id(),
        commodity,
    )
    if accept is not None and "text/csv" in accept:
        df = pd.DataFrame(output)
        return Response(
            df.to_csv(index=False, float_format="%.5f"), media_type="text/csv"
        )
    return output


@router.get("/mineral_site_location/{commodity}")
def mineral_site_location(
    commodity: str,
    accept: Annotated[str | None, Header()] = None,
):
    commodity = norm_commodity(commodity)
    output = get_mineral_site_location(
        get_snapshot_id(),
        commodity,
    )
    if accept is not None and "text/csv" in accept:
        df = pd.DataFrame(output)
        return Response(
            df.to_csv(index=False, float_format="%.5f"), media_type="text/csv"
        )
    return output


@router.get("/hyper_mineral_sites/{commodity}")
def hyper_mineral_sites(
    commodity: str,
    norm_tonnage_unit: Optional[str] = None,
    norm_grade_unit: Optional[str] = None,
    date_precision: Literal["year", "month", "day"] = "month",
    accept: Annotated[str | None, Header()] = None,
):
    commodity = norm_commodity(commodity)
    output = get_hyper_mineral_site_data(
        get_snapshot_id(),
        commodity,
        norm_tonnage_unit=norm_tonnage_unit,
        norm_grade_unit=norm_grade_unit,
        date_precision=date_precision,
    )
    if accept is not None and "text/csv" in accept:
        df = pd.DataFrame(output)
        return Response(
            df.to_csv(index=False, float_format="%.5f"), media_type="text/csv"
        )
    return output


def get_snapshot_id(endpoint=DEFAULT_ENDPOINT):
    query = "SELECT ?snapshot_id WHERE { mnr:kg dcterms:hasVersion ?snapshot_id }"
    qres = run_sparql_query(query, endpoint)
    return qres[0]["snapshot_id"]


def is_minmod_id(text: str) -> bool:
    return text.startswith("Q") and text[1:].isdigit()


def norm_commodity(commodity: str) -> str:
    if commodity.startswith("http"):
        raise HTTPException(
            status_code=404,
            detail=f"Expect commodity to be either just an id (QXXX) or name. Get `{commodity}` instead",
        )
    if not is_minmod_id(commodity):
        uri = get_commodity_by_name(commodity)
        if uri is None:
            raise HTTPException(
                status_code=404, detail=f"Commodity `{commodity}` not found"
            )
        commodity = uri
    return commodity


def get_commodity_by_name(name: str) -> Optional[str]:
    query = (
        'SELECT ?uri WHERE { ?uri a :Commodity ; rdfs:label ?name . FILTER(LCASE(STR(?name)) = "%s") }'
        % name.lower()
    )
    qres = run_sparql_query(query, DEFAULT_ENDPOINT)
    if len(qres) == 0:
        return None
    uri = qres[0]["uri"]
    assert uri.startswith(MNR_NS)
    uri = uri[len(MNR_NS) :]
    return uri


@lru_cache(maxsize=1)
def get_commodities(snapshot_id: str, endpoint=DEFAULT_ENDPOINT):
    query = """
    SELECT ?uri ?name
    WHERE {
        ?uri a :Commodity ;
            rdfs:label ?name .
    }
    """
    qres = run_sparql_query(query, endpoint)
    return qres


@lru_cache(maxsize=1)
def get_deposit_types(snapshot_id: str, endpoint=DEFAULT_ENDPOINT):
    query = """
    SELECT ?uri ?name ?environment ?group
    WHERE {
        ?uri a :DepositType ;
            rdfs:label ?name ;
            :environment ?environment ;
            :group ?group .
    }
    """
    qres = run_sparql_query(query, endpoint)
    return qres


@lru_cache(maxsize=32)
def get_hyper_mineral_site_data(
    snapshot_id: str,
    commodity: str,
    norm_tonnage_unit=None,
    norm_grade_unit=None,
    date_precision: Literal["year", "month", "day"] = "month",
    endpoint=DEFAULT_ENDPOINT,
):
    site2group = get_site_group(snapshot_id, endpoint)
    sites_info = get_mineral_site_location(
        snapshot_id=snapshot_id, commodity=commodity, endpoint=endpoint
    )
    dpt_info = get_deposit_type_classification(
        snapshot_id=snapshot_id, commodity=commodity, endpoint=endpoint
    )
    grade_tonnage_info = get_grade_tonnage_inventory(
        snapshot_id=snapshot_id,
        commodity=commodity,
        norm_tonnage_unit=norm_tonnage_unit,
        norm_grade_unit=norm_grade_unit,
        date_precision=date_precision,
        endpoint=endpoint,
    )

    # ---- 1. find top grade-tonnage per each hypersite ----
    for record in grade_tonnage_info:
        record["group_id"] = site2group[record["ms"]]
    group2best_gt = {}
    for group_id, lst in group_by_key(grade_tonnage_info, "group_id").items():
        # group by 'group_id' and find the entry with the highest 'tot_contained_metal'
        lst = [x for x in lst if x["tot_contained_metal"] is not None]
        if len(lst) == 0:
            continue
        group2best_gt[group_id] = max(lst, key=lambda x: x["tot_contained_metal"])

    # ---- 2. find top deposit type per each hypersite ----
    for record in dpt_info:
        record["group_id"] = site2group[record["ms"]]
    group2best_dpt = {}
    for group_id, lst in group_by_key(dpt_info, "group_id").items():
        # group by 'group_id' and find the entry with the highest 'top1_deposit_classification_confidence'
        lst = [
            x for x in lst if x["top1_deposit_classification_confidence"] is not None
        ]
        if len(lst) == 0:
            continue
        group2best_dpt[group_id] = max(
            lst, key=lambda x: x["top1_deposit_classification_confidence"]
        )

    # ---- 3. squash into single-row hypersites ----
    output = []
    for group_id, lst in group_by_key(sites_info, "group_id").items():
        record: dict = {"group_id": group_id}

        for key in [
            "ms",
            "ms_name",
            "ms_type",
            "ms_rank",
            "country",
            "state_or_province",
        ]:
            record[key] = sorted({x[key] for x in lst if x[key] is not None})

        record["loc_crs"] = list(
            {x["loc_crs"] for x in lst if x["loc_crs"] is not None}
        )
        record["loc_wkt"] = sorted(
            {x["loc_wkt"] for x in lst if x["loc_wkt"] is not None}
        )
        if len(record["loc_wkt"]) > 1:
            record["loc_wkt"] = merge_wkt(record["loc_wkt"])

        for key in record:
            if isinstance(record[key], list):
                if len(record[key]) == 1:
                    record[key] = record[key][0]
                elif len(record[key]) == 0:
                    record[key] = None

        for key in ["tot_contained_metal", "total_tonnage", "total_grade"]:
            if group_id not in group2best_gt:
                record[key] = None
            else:
                record[key] = group2best_gt[group_id][key]

        for key in [
            "top1_deposit_type",
            "top1_deposit_group",
            "top1_deposit_environment",
            "top1_deposit_classification_confidence",
            "top1_deposit_classification_source",
        ]:
            if group_id not in group2best_dpt:
                record[key] = None
            else:
                record[key] = group2best_dpt[group_id][key]

        output.append(record)
    return output


@lru_cache(maxsize=32)
def get_mineral_site_location(
    snapshot_id: str,
    commodity: str,
    endpoint=DEFAULT_ENDPOINT,
) -> list[dict]:
    query = (
        """
    SELECT 
        ?ms 
        ?ms_name 
        ?ms_type 
        ?ms_rank
        ?state_or_province
        ?state_or_province_name
        ?country
        ?country_name
        ?loc_crs
        ?loc_crs_name
        ?loc_wkt 
    WHERE {
        { SELECT DISTINCT ?ms WHERE { ?ms :mineral_inventory/:commodity/:normalized_uri %s . } }
        
        OPTIONAL { ?ms rdfs:label ?ms_name . }
        OPTIONAL { ?ms :site_type ?ms_type . }
        OPTIONAL { ?ms :site_rank ?ms_rank . }
        OPTIONAL { 
            ?ms :location_info ?loc .
            
            OPTIONAL { 
                ?loc :country/:normalized_uri ?country .
                ?country rdfs:label ?country_name .
            }
            OPTIONAL {
                ?loc :state_or_province/:normalized_uri ?state_or_province . 
                ?state_or_province rdfs:label ?state_or_province_name .
            }
            OPTIONAL {
                ?loc :crs/:normalized_uri ?loc_crs .
                ?loc_crs rdfs:label ?loc_crs_name .
            }
            OPTIONAL { ?loc :location ?loc_wkt . }
        }
    }
    """
        % f"mnr:{commodity}"
    )
    qres = run_sparql_query(query, endpoint)

    site2group = get_site_group(snapshot_id, endpoint)
    output = []

    for row in qres:
        record = {
            "ms": row["ms"],
            "ms_name": row["ms_name"],
            "ms_type": row["ms_type"],
            "ms_rank": row["ms_rank"],
            "country": row.get("country_name", None),
            "state_or_province": row.get("state_or_province_name", None),
            "loc_crs": row.get("loc_crs_name", None),
            "loc_wkt": row.get("loc_wkt", None),
        }

        if all(
            record[key] is None
            for key in ["country", "state_or_province", "loc_crs", "loc_wkt"]
        ):
            continue
        record["group_id"] = site2group[row["ms"]]
        output.append(record)
    return output


@lru_cache(maxsize=32)
def get_deposit_type_classification(
    snapshot_id: str,
    commodity: str,
    endpoint=DEFAULT_ENDPOINT,
):
    query = """
    SELECT
        ?ms
        ?ms_name
        ?deposit_name
        ?deposit_source
        ?deposit_confidence
        ?deposit_group
        ?deposit_environment
        ?country
        ?country_name
        ?loc_crs
        ?loc_crs_name
        ?loc_wkt
        ?state_or_province
        ?state_or_province_name
    WHERE {
        # filter by commodity
        ?ms :mineral_inventory/:commodity/:normalized_uri %s .

        # get deposit type classification
        ?ms :deposit_type_candidate ?deposit_candidate_uri .
        ?deposit_candidate_uri :source ?deposit_source .
        ?deposit_candidate_uri :confidence ?deposit_confidence .
        ?deposit_candidate_uri :normalized_uri [
            rdfs:label ?deposit_name ;
            :group ?deposit_group ;
            :environment ?deposit_environment ] .

        OPTIONAL { ?ms rdfs:label ?ms_name . }
        OPTIONAL { 
            ?ms :location_info ?loc .
            
            OPTIONAL { 
                ?loc :country/:normalized_uri ?country .
                ?country rdfs:label ?country_name .
            }
            OPTIONAL {
                ?loc :state_or_province/:normalized_uri ?state_or_province . 
                ?state_or_province rdfs:label ?state_or_province_name .
            }
            OPTIONAL {
                ?loc :crs/:normalized_uri ?loc_crs .
                ?loc_crs rdfs:label ?loc_crs_name .
            }
            OPTIONAL { ?loc :location ?loc_wkt . }
        }
    }
    """ % (
        f"mnr:{commodity}"
    )
    qres = run_sparql_query(query, endpoint)
    # Old code copy from `generate_ta2_outputs.py`
    deposits_data = pd.DataFrame(
        [
            {
                "ms": row["ms"],
                "ms_name": row["ms_name"],
                "country": row.get("country_name", None),
                "state_or_province": row.get("state_or_province_name", None),
                "loc_crs": row.get("loc_crs_name", None),
                "loc_wkt": row.get("loc_wkt", None),
                "deposit_type": row.get("deposit_name", None),
                "deposit_group": row.get("deposit_group", None),
                "deposit_environment": row.get("deposit_environment", None),
                "deposit_classification_confidence": row.get(
                    "deposit_confidence", None
                ),
                "deposit_classification_source": row.get("deposit_source", None),
            }
            for row in qres
        ]
    )

    deposits_df = deposits_data.drop_duplicates()
    deposits_df.reset_index(drop=True, inplace=True)
    deposits_df.set_index(["ms", "deposit_type"], inplace=True)
    # deposits_df['info_count'] = deposits_df[['country', 'state_or_province', 'loc_crs', 'loc_wkt']].apply(lambda x: ((x != '') & (x.notna())).sum(), axis=1)
    deposits_df["info_count"] = deposits_df[
        [
            "country",
            "state_or_province",
            "loc_crs",
            "loc_wkt",
            "deposit_classification_confidence",
        ]
    ].apply(lambda x: (((x[:-1] != "") & (x[:-1].notna())).sum(), float(x[-1])), axis=1)
    deposits_df_ordered = deposits_df.sort_values(by="info_count", ascending=False)
    deposits_df_ordered = deposits_df_ordered[
        ~deposits_df_ordered.index.duplicated(keep="first")
    ]
    deposits_df_ordered.drop(columns=["info_count"], inplace=True)
    deposits_df_ordered["deposit_classification_confidence"] = deposits_df_ordered[
        "deposit_classification_confidence"
    ].astype(float)
    deposits_df_ordered.reset_index(inplace=True)

    grouped = deposits_df_ordered.groupby("ms")
    results = []
    for name, group in grouped:
        # Sort by 'deposit_classification_confidence'
        unique_group = group.drop_duplicates(
            subset=[
                "deposit_type",
                "deposit_group",
                "deposit_environment",
                "deposit_classification_confidence",
                "deposit_classification_source",
            ]
        )
        sorted_group = unique_group.sort_values(
            "deposit_classification_confidence", ascending=False
        )
        top_5 = sorted_group.head(5)

        result_row = {"ms": name}
        for i, row in enumerate(top_5.itertuples(index=False), start=1):
            result_row[f"top{i}_deposit_type"] = row.deposit_type
            result_row[f"top{i}_deposit_group"] = row.deposit_group
            result_row[f"top{i}_deposit_environment"] = row.deposit_environment
            result_row[f"top{i}_deposit_classification_confidence"] = (
                row.deposit_classification_confidence
            )
            result_row[f"top{i}_deposit_classification_source"] = (
                row.deposit_classification_source
            )

        results.append(result_row)

    final_df = pd.DataFrame(results)
    unique_ms_data = deposits_df_ordered[
        ["ms", "ms_name", "country", "state_or_province", "loc_crs", "loc_wkt"]
    ].drop_duplicates(subset=["ms"])
    merged_deposits_df = pd.merge(unique_ms_data, final_df, on="ms", how="left")

    # fix me: temporary code to deal with nan values
    output = []
    for record in merged_deposits_df.to_dict("records"):
        for k in record:
            if pd.isna(record[k]):
                record[k] = None
        output.append(record)
    return output


@lru_cache(maxsize=32)
def get_grade_tonnage_inventory(
    snapshot_id: str,
    commodity: str,
    norm_tonnage_unit=None,
    norm_grade_unit=None,
    date_precision: Literal["year", "month", "day"] = "month",
    endpoint=DEFAULT_ENDPOINT,
):
    query = """
    SELECT 
        ?ms                       # Mineral Site URI
        ?ms_name                  # Mineral Site Name
        ?country                  # Country
        ?country_name
        ?loc_crs                  # WKT CRS
        ?loc_crs_name
        ?loc_wkt                  # WKT Geometry
        ?state_or_province
        ?state_or_province_name
        ?doc
        
        ?mi
        ?mi_cat
        ?mi_date
        ?mi_zone

        ?mi_ore_value
        ?mi_ore_unit
        ?mi_grade_value
        ?mi_grade_unit
    WHERE {
        ?ms :mineral_inventory ?mi .

        ?mi :category/:normalized_uri ?mi_cat ;
            :commodity/:normalized_uri %s ;
            :grade [
                :value ?mi_grade_value ;
                :unit/:normalized_uri ?mi_grade_unit 
            ] ;
            :ore [
                :value ?mi_ore_value ;
                :unit/:normalized_uri ?mi_ore_unit 
            ] .

        OPTIONAL { ?ms rdfs:label ?ms_name . }
        OPTIONAL { 
            ?ms :location_info ?loc .
            
            OPTIONAL { 
                ?loc :country/:normalized_uri ?country .
                ?country rdfs:label ?country_name .
            }
            OPTIONAL {
                ?loc :state_or_province/:normalized_uri ?state_or_province . 
                ?state_or_province rdfs:label ?state_or_province_name .
            }
            OPTIONAL {
                ?loc :crs/:normalized_uri ?loc_crs .
                ?loc_crs rdfs:label ?loc_crs_name .
            }
            OPTIONAL { ?loc :location ?loc_wkt . }
        }

        OPTIONAL { ?mi :reference/:document/:uri ?doc . }
        OPTIONAL { ?mi :zone ?mi_zone . }
        OPTIONAL { ?mi :date ?mi_date . }
    }
    """ % (
        f"mnr:{commodity}"
    )
    qres = run_sparql_query(
        query,
        endpoint,
        [
            "ms_name",
            "loc",
            "country",
            "country_name",
            "state_or_province",
            "state_or_province_name",
            "loc_crs",
            "loc_crs_name",
            "loc_wkt",
            "doc",
            "mi_zone",
            "mi_date",
        ],
    )
    # compute grade & tonnage for each mineral site
    grade_tonnage_model = GradeTonnageModel()

    output = []
    for ms, invs in group_by_key(qres, "ms").items():
        grade_tonnage = grade_tonnage_model(
            [
                GradeTonnageModel.MineralInventory(
                    id=inv,
                    date=inv_props[0]["mi_date"],
                    zone=inv_props[0]["mi_zone"],
                    category=[x["mi_cat"] for x in inv_props],
                    ore_value=inv_props[0]["mi_ore_value"],
                    ore_unit=inv_props[0]["mi_ore_unit"],
                    grade_value=inv_props[0]["mi_grade_value"],
                    grade_unit=inv_props[0]["mi_grade_unit"],
                )
                for inv, inv_props in group_by_key(invs, "mi").items()
            ],
            norm_tonnage_unit=norm_tonnage_unit,
            norm_grade_unit=norm_grade_unit,
        )
        record = {
            "ms": ms,
            "ms_name": invs[0]["ms_name"],
            "country": invs[0]["country_name"],
            "state_or_province": invs[0]["state_or_province_name"],
            "loc_crs": invs[0]["loc_crs_name"],
            "loc_wkt": invs[0]["loc_wkt"],
            "doc_url": invs[0]["doc"],
        }

        record.update(**fmt_grade_tonnage(grade_tonnage, norm_grade_unit))
        output.append(record)
    return output


@lru_cache(maxsize=1)
def get_site_group(snapshot_id: str, endpoint=DEFAULT_ENDPOINT):
    query = "SELECT ?s1 ?s2 WHERE { ?s1 a :MineralSite . ?s2 a :MineralSite . ?s1 owl:sameAs ?s2 . }"
    output = run_sparql_query(query, endpoint)
    G = nx.from_edgelist([(row["s1"], row["s2"]) for row in output])
    groups = nx.connected_components(G)

    mapping = {}
    for gid, group in enumerate(groups, start=1):
        for node in group:
            mapping[node] = gid

    max_group_id = max(mapping.values())

    query = "SELECT ?s1 WHERE { ?s1 a :MineralSite . FILTER NOT EXISTS { ?s1 owl:sameAs ?s2 . } }"
    output = run_sparql_query(query, DEFAULT_ENDPOINT)
    unlinked_sites = sorted(row["s1"] for row in output)
    for i, site in enumerate(unlinked_sites, start=1):
        mapping[site] = max_group_id + i

    return mapping


def fmt_grade_tonnage(
    grade_tonnage: Optional[SiteGradeTonnage], norm_grade_unit: Optional[str] = None
) -> dict:
    if grade_tonnage is None:
        grade_tonnage = SiteGradeTonnage()

    record = {
        "tot_contained_metal": grade_tonnage.total_resource_contained_metal,
        "total_tonnage": grade_tonnage.total_resource_tonnage,
        "total_grade": (grade_tonnage.get_total_resource_grade(norm_grade_unit)),
    }

    # just in case we have more data about reserve than resource
    if grade_tonnage.total_reserve_tonnage is not None and (
        grade_tonnage.total_resource_tonnage is None
        or grade_tonnage.total_reserve_tonnage > grade_tonnage.total_resource_tonnage
    ):
        record["total_tonnage"] = grade_tonnage.total_reserve_tonnage
        record["tot_contained_metal"] = grade_tonnage.total_reserve_contained_metal
        record["total_grade"] = grade_tonnage.get_total_reserve_grade(norm_grade_unit)

    return record


app.include_router(router)



from __future__ import annotations

import hashlib

from slugify import slugify

MNR_NS = "https://minmod.isi.edu/resource/"


def make_site_uri(source_id: str, record_id: str, namespace: str = MNR_NS) -> str:
    if source_id.startswith("http://"):
        if source_id.startswith("http://"):
            source_id = source_id[7:]
            if source_id.endswith("/"):
                source_id = source_id[:-1]
    elif source_id.startswith("https://"):
        source_id = source_id[8:]
        if source_id.endswith("/"):
            source_id = source_id[:-1]

    source_id = slugify(source_id)

    if isinstance(record_id, int):
        record_id = str(record_id)
    else:
        record_id = slugify(record_id)

    path = f"{source_id}__{record_id}"
    if len(path) > 120:
        path = path[:120] + "__" + hashlib.sha256(path.encode()).hexdigest()[:8]
    return f"{namespace}site__{path}"
